@@ -12,106 +12,25 @@
 
 #define REQUIRES_16BIT_INSTS __attribute__((target("16-bit-insts")))
 #define REQUIRES_GFX9_INSTS __attribute__((target("gfx9-insts")))
+#define REQUIRES_WAVE32 __attribute__((target("wavefrontsize32")))
+#define REQUIRES_WAVE64 __attribute__((target("wavefrontsize64")))
 
 // Generic intrinsics
 extern __attribute__((const)) float2 __llvm_fma_2f32(float2, float2, float2) __asm("llvm.fma.v2f32");
 
 extern __attribute__((const)) half2 __llvm_fma_2f16(half2, half2, half2) __asm("llvm.fma.v2f16");
-extern __attribute__((const)) half2 __llvm_fabs_2f16(half2) __asm("llvm.fabs.v2f16");
-extern __attribute__((const)) half2 __llvm_minnum_2f16(half2, half2) __asm("llvm.minnum.v2f16");
-extern __attribute__((const)) half2 __llvm_maxnum_2f16(half2, half2) __asm("llvm.maxnum.v2f16");
-extern __attribute__((const)) half2 __llvm_copysign_2f16(half2, half2) __asm("llvm.copysign.v2f16");
-extern __attribute__((const)) half2 __llvm_floor_2f16(half2) __asm("llvm.floor.v2f16");
-extern __attribute__((const)) half2 __llvm_ceil_2f16(half2) __asm("llvm.ceil.v2f16");
-extern __attribute__((const)) half2 __llvm_trunc_2f16(half2) __asm("llvm.trunc.v2f16");
 extern __attribute__((const)) half2 __llvm_round_2f16(half2) __asm("llvm.round.v2f16");
 extern __attribute__((const)) half2 __llvm_rint_2f16(half2) __asm("llvm.rint.v2f16");
-extern __attribute__((const)) half2 __llvm_canonicalize_2f16(half2) __asm("llvm.canonicalize.v2f16");
 
-// Intrinsics requiring wrapping
-extern __attribute__((const)) uchar __llvm_ctlz_i8(uchar);
-extern __attribute__((const)) ushort __llvm_ctlz_i16(ushort);
-extern __attribute__((const)) uint __llvm_ctlz_i32(uint);
-extern __attribute__((const)) ulong __llvm_ctlz_i64(ulong);
+#define BUILTIN_CLZ_U8(x) (uchar)(x == 0u ? 8 : __builtin_clz(x) - 24)
+#define BUILTIN_CLZ_U16(x) (ushort)(x == 0u ? 16 : __builtin_clzs(x))
+#define BUILTIN_CLZ_U32(x) (uint)(x == 0u ? 32 : __builtin_clz(x))
+#define BUILTIN_CLZ_U64(x) (ulong)(x == 0u ? 64 : __builtin_clzl(x))
 
-extern __attribute__((const)) uchar __llvm_cttz_i8(uchar);
-extern __attribute__((const)) ushort __llvm_cttz_i16(ushort);
-extern __attribute__((const)) uint __llvm_cttz_i32(uint);
-extern __attribute__((const)) ulong __llvm_cttz_i64(ulong);
-
-// Atomics
-extern uint __llvm_ld_atomic_a1_x_dev_i32(__global uint *);
-extern ulong __llvm_ld_atomic_a1_x_dev_i64(__global ulong *);
-extern uint __llvm_ld_atomic_a3_x_wg_i32(__local uint *);
-extern ulong __llvm_ld_atomic_a3_x_wg_i64(__local ulong *);
-
-extern void __llvm_st_atomic_a1_x_dev_i32(__global uint *, uint);
-extern void __llvm_st_atomic_a1_x_dev_i64(__global ulong *, ulong);
-extern void __llvm_st_atomic_a3_x_wg_i32(__local uint *, uint);
-extern void __llvm_st_atomic_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_atomic_add_a1_x_dev_i32(__global uint *, uint);
-extern ulong __llvm_atomic_add_a1_x_dev_i64(__global ulong *, ulong);
-extern uint __llvm_atomic_add_a3_x_wg_i32(__local uint *, uint);
-extern ulong __llvm_atomic_add_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_atomic_and_a1_x_dev_i32(__global uint *, uint);
-extern ulong __llvm_atomic_and_a1_x_dev_i64(__global ulong *, ulong);
-extern uint __llvm_atomic_and_a3_x_wg_i32(__local uint *, uint);
-extern ulong __llvm_atomic_and_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_atomic_or_a1_x_dev_i32(__global uint *, uint);
-extern ulong __llvm_atomic_or_a1_x_dev_i64(__global ulong *, ulong);
-extern uint __llvm_atomic_or_a3_x_wg_i32(__local uint *, uint);
-extern ulong __llvm_atomic_or_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_atomic_max_a1_x_dev_i32(__global int *, int);
-extern uint __llvm_atomic_umax_a1_x_dev_i32(__global uint *, uint);
-extern ulong __llvm_atomic_max_a1_x_dev_i64(__global long *, long);
-extern ulong __llvm_atomic_umax_a1_x_dev_i64(__global ulong *, ulong);
-extern uint __llvm_atomic_max_a3_x_wg_i32(__local int *, int);
-extern uint __llvm_atomic_umax_a3_x_wg_i32(__local uint *, uint);
-extern ulong __llvm_atomic_max_a3_x_wg_i64(__local long *, long);
-extern ulong __llvm_atomic_umax_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_atomic_min_a1_x_dev_i32(__global int *, int);
-extern uint __llvm_atomic_umin_a1_x_dev_i32(__global uint *, uint);
-extern ulong __llvm_atomic_min_a1_x_dev_i64(__global long *, long);
-extern ulong __llvm_atomic_umin_a1_x_dev_i64(__global ulong *, ulong);
-extern uint __llvm_atomic_min_a3_x_wg_i32(__local int *, int);
-extern uint __llvm_atomic_umin_a3_x_wg_i32(__local uint *, uint);
-extern ulong __llvm_atomic_min_a3_x_wg_i64(__local long *, long);
-extern ulong __llvm_atomic_umin_a3_x_wg_i64(__local ulong *, ulong);
-
-extern uint __llvm_cmpxchg_a1_x_x_dev_i32(__global uint *, uint, uint);
-extern ulong __llvm_cmpxchg_a1_x_x_dev_i64(__global ulong *, ulong, ulong);
-extern uint __llvm_cmpxchg_a3_x_x_wg_i32(__local uint *, uint, uint);
-extern ulong __llvm_cmpxchg_a3_x_x_wg_i64(__local ulong *, ulong, ulong);
-
-// AMDGPU intrinsics
-
-// llvm.amdgcn.mov.dpp.i32 <src> <dpp_ctrl> <row_mask> <bank_mask> <bound_ctrl>
-
-// llvm.amdgcn.update.dpp.i32 <old> <src> <dpp_ctrl> <row_mask> <bank_mask> <bound_ctrl>
-extern uint __llvm_amdgcn_update_dpp_i32(uint, uint, uint, uint, uint, bool) __asm("llvm.amdgcn.update.dpp.i32");
-
-// llvm.amdgcn.mov.dpp8.i32 <src> <sel>
-extern uint __llvm_amdgcn_dpp8_i32(uint, uint) __asm("llvm.amdgcn.dpp8.i32");
-
-// llvm.amdgcn.permlane16 <old> <src0> <src1> <src2> <fi> <bound_control>
-extern uint __llvm_amdgcn_permlane16(uint, uint, uint, uint, bool, bool) __asm("llvm.amdgcn.permlane16");
-
-// llvm.amdgcn.permlanex16 <old> <src0> <src1> <src2> <fi> <bound_control>
-extern uint __llvm_amdgcn_permlanex16(uint, uint, uint, uint, bool, bool) __asm("llvm.amdgcn.permlanex16");
-
-extern __attribute__((const, convergent)) ulong __llvm_amdgcn_icmp_i64_i32(uint, uint, uint) __asm("llvm.amdgcn.icmp.i64.i32");
-extern __attribute__((const, convergent)) ulong __llvm_amdgcn_icmp_i64_i64(ulong, ulong, uint) __asm("llvm.amdgcn.icmp.i64.i64");
-extern __attribute__((const, convergent)) ulong __llvm_amdgcn_fcmp_i64_f32(float, float, uint) __asm("llvm.amdgcn.fcmp.i64.f32");
-extern __attribute__((const, convergent)) ulong __llvm_amdgcn_fcmp_i64_f64(double, double, uint) __asm("llvm.amdgcn.fcmp.i64.f64");
-extern __attribute__((const, convergent)) uint __llvm_amdgcn_icmp_i32_i32(uint, uint, uint) __asm("llvm.amdgcn.icmp.i32.i32");
-extern __attribute__((const, convergent)) uint __llvm_amdgcn_icmp_i32_i64(ulong, ulong, uint) __asm("llvm.amdgcn.icmp.i32.i64");
-extern __attribute__((const, convergent)) uint __llvm_amdgcn_fcmp_i32_f32(float, float, uint) __asm("llvm.amdgcn.fcmp.i32.f32");
-extern __attribute__((const, convergent)) uint __llvm_amdgcn_fcmp_i32_f64(double, double, uint) __asm("llvm.amdgcn.fcmp.i32.f64");
+#define BUILTIN_CTZ_U8(x) (uchar)(x == 0u ? (uchar)8 : __builtin_ctz((uint)x))
+#define BUILTIN_CTZ_U16(x) (ushort)(x == 0u ? 16 : __builtin_ctzs(x))
+#define BUILTIN_CTZ_U32(x) (uint)(x == 0u ? 32 : __builtin_ctz(x))
+#define BUILTIN_CTZ_U64(x) (ulong)(x == 0u ? 64 : __builtin_ctzl(x))
 
 // Buffer Load/Store
 extern __attribute__((pure)) float4 __llvm_amdgcn_struct_buffer_load_format_v4f32(uint4 rsrc, uint vindex, uint voffset, uint soffset, uint cachepolicy) __asm("llvm.amdgcn.struct.buffer.load.format.v4f32");
