@@ -39,14 +39,14 @@
 #define BUILTIN_ABS_F32 __builtin_fabsf
 #define BUILTIN_ABS_F64 __builtin_fabs
 #define BUILTIN_ABS_F16 __builtin_fabsf16
-#define BUILTIN_ABS_2F16 __llvm_fabs_2f16
+#define BUILTIN_ABS_2F16 __builtin_elementwise_abs
 
 #define BUILTIN_BITALIGN_B32 __builtin_amdgcn_alignbit
 
 #define BUILTIN_CEIL_F32 __builtin_ceilf
 #define BUILTIN_CEIL_F64 __builtin_ceil
 #define BUILTIN_CEIL_F16 __builtin_ceilf16
-#define BUILTIN_CEIL_2F16 __llvm_ceil_2f16
+#define BUILTIN_CEIL_2F16 __builtin_elementwise_ceil
 
 #define BUILTIN_CLASS_F32 __builtin_amdgcn_classf
 #define BUILTIN_CLASS_F64 __builtin_amdgcn_class
@@ -68,35 +68,44 @@
 #define BUILTIN_ISFINITE_F64(x) __builtin_amdgcn_class(x, CLASS_NNOR|CLASS_NSUB|CLASS_NZER|CLASS_PZER|CLASS_PSUB|CLASS_PNOR)
 #define BUILTIN_ISFINITE_F16(x) __builtin_amdgcn_classh(x, CLASS_NNOR|CLASS_NSUB|CLASS_NZER|CLASS_PZER|CLASS_PSUB|CLASS_PNOR)
 
+#define BUILTIN_ISSUBNORMAL_F32(x) __builtin_amdgcn_classf(x, CLASS_NSUB|CLASS_PSUB)
+#define BUILTIN_ISSUBNORMAL_F64(x) __builtin_amdgcn_class(x, CLASS_NSUB|CLASS_PSUB)
+#define BUILTIN_ISSUBNORMAL_F16(x) __builtin_amdgcn_classh(x, CLASS_NSUB|CLASS_PSUB)
+
+#define BUILTIN_ISZERO_F32(x) __builtin_amdgcn_classf(x, CLASS_NZER|CLASS_PZER)
+#define BUILTIN_ISZERO_F64(x) __builtin_amdgcn_class(x, CLASS_NZER|CLASS_PZER)
+#define BUILTIN_ISZERO_F16(x) __builtin_amdgcn_classh(x, CLASS_NZER|CLASS_PZER)
+
+#define BUILTIN_ISNORMAL_F32(x) __builtin_amdgcn_classf(x, CLASS_NNOR|CLASS_PNOR)
+#define BUILTIN_ISNORMAL_F64(x) __builtin_amdgcn_class(x, CLASS_NNOR|CLASS_PNOR)
+#define BUILTIN_ISNORMAL_F16(x) __builtin_amdgcn_classh(x, CLASS_NNOR|CLASS_PNOR)
+
 #define BUILTIN_COPYSIGN_F32 __builtin_copysignf
 #define BUILTIN_COPYSIGN_F64 __builtin_copysign
 #define BUILTIN_COPYSIGN_F16 __builtin_copysignf16
-#define BUILTIN_COPYSIGN_2F16 __llvm_copysign_2f16
-
-#define BUILTIN_CLZ_U32 __llvm_ctlz_i32
-#define BUILTIN_CLZ_U64 __llvm_ctlz_i64
+#define BUILTIN_COPYSIGN_2F16 __builtin_elementwise_copysign
 
 #define BUILTIN_FLOOR_F32 __builtin_floorf
 #define BUILTIN_FLOOR_F64 __builtin_floor
 #define BUILTIN_FLOOR_F16 __builtin_floorf16
-#define BUILTIN_FLOOR_2F16 __llvm_floor_2f16
+#define BUILTIN_FLOOR_2F16 __builtin_elementwise_floor
 
 #define BUILTIN_FRACTION_F32(X) ({ \
     float _fract_x = X; \
     float _fract_r = __builtin_amdgcn_fractf(_fract_x); \
-    _fract_r = __builtin_amdgcn_classf(_fract_x, CLASS_PINF|CLASS_NINF) ? 0.0f : _fract_r; \
+    _fract_r = BUILTIN_ISINF_F32(_fract_x) ? 0.0f : _fract_r; \
     _fract_r; \
 })
 #define BUILTIN_FRACTION_F64(X) ({ \
     double _fract_x = X; \
     double _fract_r = __builtin_amdgcn_fract(_fract_x); \
-    _fract_r = __builtin_amdgcn_class(_fract_x, CLASS_PINF|CLASS_NINF) ? 0.0 : _fract_r; \
+    _fract_r = BUILTIN_ISINF_F64(_fract_x) ? 0.0 : _fract_r; \
     _fract_r; \
 })
 #define BUILTIN_FRACTION_F16(X) ({ \
     half _fract_x = X; \
     half _fract_r = __builtin_amdgcn_fracth(_fract_x); \
-    _fract_r = __builtin_amdgcn_classh(_fract_x, CLASS_PINF|CLASS_NINF) ? 0.0h : _fract_r; \
+    _fract_r = BUILTIN_ISINF_F16(_fract_x) ? 0.0h : _fract_r; \
     _fract_r; \
 })
 
@@ -105,7 +114,7 @@
 #define BUILTIN_MAX_F32 __builtin_fmaxf
 #define BUILTIN_MAX_F64 __builtin_fmax
 #define BUILTIN_MAX_F16 __builtin_fmaxf16
-#define BUILTIN_MAX_2F16 __llvm_maxnum_2f16
+#define BUILTIN_MAX_2F16 __builtin_elementwise_max
 
 #define BUILTIN_MAX_S32(A,B) ((A) < (B) ? (B) : (A))
 #define BUILTIN_MAX_U32(A,B) ((A) < (B) ? (B) : (A))
@@ -113,7 +122,7 @@
 #define BUILTIN_MIN_F32 __builtin_fminf
 #define BUILTIN_MIN_F64 __builtin_fmin
 #define BUILTIN_MIN_F16 __builtin_fminf16
-#define BUILTIN_MIN_2F16 __llvm_minnum_2f16
+#define BUILTIN_MIN_2F16 __builtin_elementwise_min
 
 #define BUILTIN_MIN_S32(A,B) ((A) < (B) ? (A) : (B))
 #define BUILTIN_MIN_U32(A,B) ((A) < (B) ? (A) : (B))
@@ -121,7 +130,6 @@
 #define BUILTIN_CANONICALIZE_F32(X) __builtin_canonicalizef(X)
 #define BUILTIN_CANONICALIZE_F64(X) __builtin_canonicalize(X)
 #define BUILTIN_CANONICALIZE_F16(X) __builtin_canonicalizef16(X)
-#define BUILTIN_CANONICALIZE_2F16(X) __llvm_canonicalize_2f16(X)
 
 #define BUILTIN_MULHI_U32(A,B) (((ulong)(A) * (ulong)(B)) >> 32)
 
@@ -155,7 +163,7 @@
 #define BUILTIN_TRUNC_F32 __builtin_truncf
 #define BUILTIN_TRUNC_F64 __builtin_trunc
 #define BUILTIN_TRUNC_F16 __builtin_truncf16
-#define BUILTIN_TRUNC_2F16 __llvm_trunc_2f16
+#define BUILTIN_TRUNC_2F16 __builtin_elementwise_trunc
 
 #define BUILTIN_ROUND_F32 __builtin_roundf
 #define BUILTIN_ROUND_F64 __builtin_round
@@ -204,12 +212,12 @@
 #define BUILTIN_CMAX_F32 __builtin_fmaxf
 #define BUILTIN_CMAX_F64 __builtin_fmax
 #define BUILTIN_CMAX_F16 __builtin_fmaxf16
-#define BUILTIN_CMAX_2F16 __llvm_maxnum_2f16
+#define BUILTIN_CMAX_2F16 __builtin_elementwise_max
 
 #define BUILTIN_CMIN_F32 __builtin_fminf
 #define BUILTIN_CMIN_F64 __builtin_fmin
 #define BUILTIN_CMIN_F16 __builtin_fminf16
-#define BUILTIN_CMIN_2F16 __llvm_minnum_2f16
+#define BUILTIN_CMIN_2F16 __builtin_elementwise_min
 
 #define BUILTIN_TRIG_PREOP_F64 __builtin_amdgcn_trig_preop
 
@@ -232,81 +240,12 @@
 #define BUILTIN_CLAMP_F32(X,L,H) __builtin_amdgcn_fmed3f(X,L,H)
 #define BUILTIN_CLAMP_F16(X,L,H) __llvm_amdgcn_fmed3_f16(X,L,H)
 
-#define BUILTIN_ADD_RTE_F32 __llvm_add_rte_f32
-#define BUILTIN_ADD_RTE_F64 __llvm_add_rte_f64
-#define BUILTIN_ADD_RTE_F16 __llvm_add_rte_f16
-#define BUILTIN_ADD_RTN_F32 __llvm_add_rtn_f32
-#define BUILTIN_ADD_RTN_F64 __llvm_add_rtn_f64
-#define BUILTIN_ADD_RTN_F16 __llvm_add_rtn_f16
-#define BUILTIN_ADD_RTP_F32 __llvm_add_rtp_f32
-#define BUILTIN_ADD_RTP_F64 __llvm_add_rtp_f64
-#define BUILTIN_ADD_RTP_F16 __llvm_add_rtp_f16
-#define BUILTIN_ADD_RTZ_F32 __llvm_add_rtz_f32
-#define BUILTIN_ADD_RTZ_F64 __llvm_add_rtz_f64
-#define BUILTIN_ADD_RTZ_F16 __llvm_add_rtz_f16
-
-#define BUILTIN_SUB_RTE_F32 __llvm_sub_rte_f32
-#define BUILTIN_SUB_RTE_F64 __llvm_sub_rte_f64
-#define BUILTIN_SUB_RTE_F16 __llvm_sub_rte_f16
-#define BUILTIN_SUB_RTN_F32 __llvm_sub_rtn_f32
-#define BUILTIN_SUB_RTN_F64 __llvm_sub_rtn_f64
-#define BUILTIN_SUB_RTN_F16 __llvm_sub_rtn_f16
-#define BUILTIN_SUB_RTP_F32 __llvm_sub_rtp_f32
-#define BUILTIN_SUB_RTP_F64 __llvm_sub_rtp_f64
-#define BUILTIN_SUB_RTP_F16 __llvm_sub_rtp_f16
-#define BUILTIN_SUB_RTZ_F32 __llvm_sub_rtz_f32
-#define BUILTIN_SUB_RTZ_F64 __llvm_sub_rtz_f64
-#define BUILTIN_SUB_RTZ_F16 __llvm_sub_rtz_f16
-
-#define BUILTIN_MUL_RTE_F32 __llvm_mul_rte_f32
-#define BUILTIN_MUL_RTE_F64 __llvm_mul_rte_f64
-#define BUILTIN_MUL_RTE_F16 __llvm_mul_rte_f16
-#define BUILTIN_MUL_RTN_F32 __llvm_mul_rtn_f32
-#define BUILTIN_MUL_RTN_F64 __llvm_mul_rtn_f64
-#define BUILTIN_MUL_RTN_F16 __llvm_mul_rtn_f16
-#define BUILTIN_MUL_RTP_F32 __llvm_mul_rtp_f32
-#define BUILTIN_MUL_RTP_F64 __llvm_mul_rtp_f64
-#define BUILTIN_MUL_RTP_F16 __llvm_mul_rtp_f16
-#define BUILTIN_MUL_RTZ_F32 __llvm_mul_rtz_f32
-#define BUILTIN_MUL_RTZ_F64 __llvm_mul_rtz_f64
-#define BUILTIN_MUL_RTZ_F16 __llvm_mul_rtz_f16
-
-#define BUILTIN_DIV_RTE_F32 __llvm_div_rte_f32
-#define BUILTIN_DIV_RTE_F64 __llvm_div_rte_f64
-#define BUILTIN_DIV_RTE_F16 __llvm_div_rte_f16
-#define BUILTIN_DIV_RTN_F32 __llvm_div_rtn_f32
-#define BUILTIN_DIV_RTN_F64 __llvm_div_rtn_f64
-#define BUILTIN_DIV_RTN_F16 __llvm_div_rtn_f16
-#define BUILTIN_DIV_RTP_F32 __llvm_div_rtp_f32
-#define BUILTIN_DIV_RTP_F64 __llvm_div_rtp_f64
-#define BUILTIN_DIV_RTP_F16 __llvm_div_rtp_f16
-#define BUILTIN_DIV_RTZ_F32 __llvm_div_rtz_f32
-#define BUILTIN_DIV_RTZ_F64 __llvm_div_rtz_f64
-#define BUILTIN_DIV_RTZ_F16 __llvm_div_rtz_f16
-
-#define BUILTIN_SQRT_RTE_F32 __llvm_sqrt_rte_f32
-#define BUILTIN_SQRT_RTE_F64 __llvm_sqrt_rte_f64
-#define BUILTIN_SQRT_RTE_F16 __llvm_sqrt_rte_f16
-#define BUILTIN_SQRT_RTN_F32 __llvm_sqrt_rtn_f32
-#define BUILTIN_SQRT_RTN_F64 __llvm_sqrt_rtn_f64
-#define BUILTIN_SQRT_RTN_F16 __llvm_sqrt_rtn_f16
-#define BUILTIN_SQRT_RTP_F32 __llvm_sqrt_rtp_f32
-#define BUILTIN_SQRT_RTP_F64 __llvm_sqrt_rtp_f64
-#define BUILTIN_SQRT_RTP_F16 __llvm_sqrt_rtp_f16
-#define BUILTIN_SQRT_RTZ_F32 __llvm_sqrt_rtz_f32
-#define BUILTIN_SQRT_RTZ_F64 __llvm_sqrt_rtz_f64
-#define BUILTIN_SQRT_RTZ_F16 __llvm_sqrt_rtz_f16
-
-#define BUILTIN_FMA_RTE_F32 __llvm_fma_rte_f32
-#define BUILTIN_FMA_RTE_F64 __llvm_fma_rte_f64
-#define BUILTIN_FMA_RTE_F16 __llvm_fma_rte_f16
-#define BUILTIN_FMA_RTN_F32 __llvm_fma_rtn_f32
-#define BUILTIN_FMA_RTN_F64 __llvm_fma_rtn_f64
-#define BUILTIN_FMA_RTN_F16 __llvm_fma_rtn_f16
-#define BUILTIN_FMA_RTP_F32 __llvm_fma_rtp_f32
-#define BUILTIN_FMA_RTP_F64 __llvm_fma_rtp_f64
-#define BUILTIN_FMA_RTP_F16 __llvm_fma_rtp_f16
-#define BUILTIN_FMA_RTZ_F32 __llvm_fma_rtz_f32
-#define BUILTIN_FMA_RTZ_F64 __llvm_fma_rtz_f64
-#define BUILTIN_FMA_RTZ_F16 __llvm_fma_rtz_f16
+#define ROUND_RTE 0
+#define ROUND_RTP 1
+#define ROUND_RTN 2
+#define ROUND_RTZ 3
 
+#define BUILTIN_GETROUND_F32() __builtin_amdgcn_s_getreg((1 << 0) | (0 << 6) | ((2-1) << 11))
+#define BUILTIN_SETROUND_F32(X) __builtin_amdgcn_s_setreg((1 << 0) | (0 << 6) | ((2-1) << 11), X)
+#define BUILTIN_GETROUND_F16F64() __builtin_amdgcn_s_getreg((1 << 0) | (2 << 6) | ((2-1) << 11))
+#define BUILTIN_SETROUND_F16F64(X) __builtin_amdgcn_s_setreg((1 << 0) | (2 << 6) | ((2-1) << 11), X)
